//********************************************************************
//* HHE-142 ESP32 Internet Radio
//* 2023年5月5日 May 5, 2023
//* Version 1.00 by JG1CCL W3CCL Beard UCHIDA CCLab(Rabbit Hat Design)
//********************************************************************
//*  Uses MAX98357 I2S Amplifier Module
//*  Uses ESP32-S3(ESP32-1732S019)
//*  Uses ESP32-audioI2S Library
//*    https://github.com/schreibfaul1/ESP32-audioI2S
//*  Uses LovyanGFX Library
//*    https://github.com/lovyan03/LovyanGFX
//*  Uses buttonarray Internal Library (contributed by W6IPA)
//*    https://github.com/k3ng/k3ng_cw_keyer
//*    I would like to thank W6IPA.
//********************************************************************

//********************************************************************
//* インクルード（Include）
//********************************************************************
// Include required libraries
#define LGFX_USE_V1
#include <LovyanGFX.hpp>
#include "LGFX_ESP32-S3_ESP32-1732S019.hpp"

#include <Arduino.h>
#include <WiFi.h>
#include <Audio.h>
#include <FS.h>
#include <SPI.h>
#include "src/buttonarray/buttonarray.h"

//********************************************************************
//* 定義（Definition）
//********************************************************************
#define DEBUG

// Define SPI connections
// SPIClass SPIsd(HSPI);
#define SD_SS   18
#define SD_MOSI 17
#define SD_SCK  16
#define SD_MISO 15

// Define I2S connections
#define I2S_LRC  40
#define I2S_BCLK 41
#define I2S_DOUT 42

// Define ButtonArray
#define ARRAY_PIN   4
#define NB_BUTTONS  5
//#define ENCODER_PIN 5

// Define Operation related
#define SL_SHORT_INCREMENT 1      // Station list index short increment
#define SL_LONG_INCREMENT 10      // Station list index long increment
#define VOL_SHORT_INCREMENT 1     // Volume short increment
#define VOL_LONG_INCREMENT 10     // Volume long increment
#define STATION_LIST_INDEX_MIN  0 // Minimum value of station list
#define STATION_LIST_INDEX_MAX 99 // Maximum value of station list
#define DEFAULT_VOLUME 5          // Default volume level
#define MIN_VOLUME     0          // Minimum volume level
#define MAX_VOLUME   100          // Maximum volume level

//********************************************************************
//* 変数宣言（ Variable Declaration）
//********************************************************************
LGFX panel;   // Create panel object
Audio audio;  // Create audio object
ButtonArray button_array(ARRAY_PIN, NB_BUTTONS, false); // Create button_array object

// Wifi Credentials
String ssid =    "SSID?";       // SSID
String password = "Password?";  // Password

// Station list related
char StationListFileName[] = "/Station.m3u";  // Station list file name
int max_i;  // Maxumum station number
struct StationData{
  char  StationName[128], URL[128];
} StationList[STATION_LIST_INDEX_MAX + 1];  // Station data

//********************************************************************
//* setup()
//********************************************************************
void setup(void) {
  int i;  // Counter variable

  // Start Serial Monitor
  Serial.begin(115200);

  // LovyanGFX Initial screen display
  panel.init();
  panel.startWrite();
  panel.setRotation(1);
  panel.fillScreen(TFT_BLACK);
  panel.setFont(&fonts::Font2); // H8 * V16
  panel.setTextSize(2, 1.5);    // H16 * V24
  panel.setTextColor(TFT_WHITE);
  panel.println("HHE-142");
  panel.println(" ESP32 Internet Radio");
  panel.println("  Version 1.00");
  panel.println("   by JG1CCL W3CCL");
  panel.println("    Beard UCHIDA");
  panel.println("    CCLab");
  panel.println("    (Rabbit Hat Design)");
  //display.endWrite();
  delay(3000);

  // Setup WiFi in Station mode
  WiFi.disconnect();
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid.c_str(), password.c_str());
  // Repeat until connected
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  // WiFi Connected, print IP to serial monitor
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.println("-----");

  // Set NTP server and local time zone (specify offset)
  configTime(9 * 3600L, 0, "ntp.nict.jp", "time.google.com", "ntp.jst.mfeed.ad.jp");

  // Connect MAX98357 I2S Amplifier Module
  audio.setPinout(I2S_BCLK, I2S_LRC, I2S_DOUT);

  // SD Card
  //SPIsd.begin(SD_SCK, SD_MISO, SD_MOSI, SD_SS);  // SPI3_HOST(HSPI) SCK、MISO、MOSI、SS
  //if (!SD.begin(SD_SS, SPIsd)) {
  SPI.begin(SD_SCK, SD_MISO, SD_MOSI, SD_SS);  // SPI2_HOST(VSPI) SCK、MISO、MOSI、SS
  if (!SD.begin(SD_SS, SPI)) {
  Serial.println("Card Mount Failed");
    return;
  }
  uint8_t cardType = SD.cardType();
  if (cardType == CARD_NONE) {
  Serial.println("No SD card attached");
    return;
  }
  Serial.print("SD Card Type: ");
  if (cardType == CARD_MMC) {
    Serial.println("MMC");
  } else if (cardType == CARD_SD) {
    Serial.println("SDSC");
  } else if (cardType == CARD_SDHC) {
    Serial.println("SDHC");
  } else {
    Serial.println("UNKNOWN");
  }
  uint64_t cardSize = SD.cardSize() / (1024 * 1024);
  Serial.printf("SD Card Size: %lluMB\n", cardSize);
  max_i = readStationData(SD, StationListFileName);
  Serial.printf("%02d\n", max_i);
  for (i = 0; i < max_i; i++) {
    Serial.printf("%02d\n", i);
    Serial.println(StationList[i].StationName);
    Serial.println(StationList[i].URL);
  }

  // パネルに表示し局へ接続する（Display on panel and connect to station）
  panelConnectToHost(StationList[0].StationName, StationList[0].URL);

  // 音量を設定しパネルに表示する（Volume is set and displayed on the panel）
  panelSetVolume(DEFAULT_VOLUME);

  // Creation of judgment criteria table
  button_array.AddAll();
}

//********************************************************************
//* loop()
//********************************************************************
void loop(void) {
  uint32_t voltage;
  uint32_t raw, araw;
  static int i = STATION_LIST_INDEX_MIN, j = DEFAULT_VOLUME;
  static int menubutton = 0;
  int ii; // Station list index increment
  int ji; // Station name or URL index increment
  long button_depress_time;
  int analogbutton;

  // ボタンを検査する（Inspect for button presses）
  analogbutton = button_array.Pressed();
 
  // ボタンを押したときには（When the button is pressed）
  if (analogbutton >= 0) {
    // 演奏を停止する（Stop playing）
    //audio.stopSong();
    // 最後に押した時間を取得する（Get the time of the last press）
    button_depress_time = button_array.last_pressed_ms;

    // ボタンを離すまで繰り返す（Repeat until the button is released）
    //while (button_array.Held(analogbutton, button_depress_time + 1000)) {}
    while (button_array.Held(analogbutton)) {}

    // 短押のときには（Push short button）
    if (millis() - button_depress_time < 400) {
      #if defined(DEBUG)
        Serial.println("Button short pressed");
        Serial.println(millis() - button_depress_time);
      #endif
      ii = SL_SHORT_INCREMENT;
      ji = VOL_SHORT_INCREMENT;
    
    // それ以外（長押）のときには（Push long button）
    } else {
      #if defined(DEBUG)
        Serial.println("Button long pressed");
        Serial.println(millis() - button_depress_time);
      #endif
      ii = SL_LONG_INCREMENT;
      ji = VOL_LONG_INCREMENT;
    }

    // ボタン種類により分岐する（Branching by button type）
    switch (analogbutton) {
      // Button1 <ch(Previous)
      case 0:
        #if defined(DEBUG)
          Serial.println("Button1 <ch(Previous) pressed");
        #endif
        i = i - ii; // Station list index decrement
        // 最小値を超えたときには最小値にする（Minimum value when minimum value is exceeded）
        if (i < STATION_LIST_INDEX_MIN) i = STATION_LIST_INDEX_MIN;
        break;

      // Button2 Vol-
      case 1:
        #if defined(DEBUG)
          Serial.println("Button2 Vol- pressed");
        #endif
        j = j - ji; // Volume decrement

        // 最小値を超えたときには最小値にする（Minimum value when minimum value is exceeded）
        if (j < MIN_VOLUME) j = MIN_VOLUME;
        #if defined(DEBUG)
          Serial.printf("Volume %d\n", j);
        #endif
        break;

      // Button3 Menu
      case 2:
        #if defined(DEBUG)
          Serial.println("Button3 Menu pressed");
        #endif
        if (menubutton != 0) {
          #if defined(DEBUG)
            Serial.printf("max_i Ch %d%d\n", max_i, i);
          #endif
          // パネルに表示し局へ接続する（Display on panel and connect to station）
          panelConnectToHost(StationList[i].StationName, StationList[i].URL);

          menubutton = 0;
        } else {
          menubutton = ji;
        }
        break;

      // Button4 Vol+
      case 3:
        #if defined(DEBUG)
          Serial.println("Button4 Vol+ pressed");
        #endif
        j = j + ji; // Volume increment
        // 最大値を超えたときには最大値にする（Maximize when the maximum value is exceeded）
        if (j > MAX_VOLUME) j = MAX_VOLUME;
        #if defined(DEBUG)
          Serial.printf("Volume %d\n", j);
        #endif
        break;

      // Button5 ch>(Next)
      case 4:
        #if defined(DEBUG)
          Serial.println("Button5 ch>(Next) pressed");
        #endif
        i = i + ii; // Station list index increment
        // 最大値を超えたときには にする（Maximize when the maximum value is exceeded）
        if (i > max_i) i = max_i;
        break;

      // Default
      default:
        #if defined(DEBUG)
          Serial.println("Default pressed");
        #endif
        break;
    }

    //
    if (analogbutton == 0 || analogbutton ==4) {
      #if defined(DEBUG)
        Serial.printf("max_i %d Ch %d\n", max_i, i);
      #endif
      // パネルに表示し局へ接続する（Display on panel and connect to station）
      panelConnectToHost(StationList[i].StationName, StationList[i].URL);
    }

    #if defined(DEBUG)
      Serial.printf("Volume %d\n", j);
    #endif
    // 音量を設定しパネルに表示する（Volume is set and displayed on the panel）
    panelSetVolume(j);

  //
  } else if (menubutton != 0) {
    // パネルにローカル時間を表示する（Display local time on panel）
    panelLoaclTime();
    delay(1000);

  // それ以外のときには
  } else {
    // 演奏する（Run audio player）
    audio.loop();
  }
}

//********************************************************************
//* 関数（Functions）
//********************************************************************
//*
//* 音量を設定しパネルに表示する（Volume is set and displayed on the panel）
//*
void panelSetVolume(int volume) {
  panel.fillRect (0, 146, 320, 24, TFT_BLACK);
  panel.setCursor(0, 146);
  panel.setTextColor(TFT_BLUE);
  panel.printf("Volume %d", volume);
  // 音量を設定する（Set volume）
  // Set thevolume (0-100)
  audio.setVolume(volume);
}

//*
//* パネルに表示し局へ接続する（Display on panel and connect to station）
//*
void panelConnectToHost(char stationname[], char url[]) {
  panel.fillScreen(TFT_BLACK);
  panel.setCursor(0, 0);
  panel.setFont(&fonts::Font2); // H8 * V16
  panel.setTextSize(2, 1.5);    // H16 * V24
  panel.setTextColor(TFT_RED);
  panel.print(stationname);
  // 局へ接続する（Connect to host）
  audio.connecttohost(url);
}

//*
//* パネルにローカル時間を表示する（Display local time on panel）
//*
void panelLoaclTime(void) {
  struct tm timeInfo;

  // ローカル時間を取得する（Get local time）
  getLocalTime(&timeInfo);
  #if defined(DEBUG)
    Serial.printf(" %04d/%02d/%02d %02d:%02d:%02d\n",
      timeInfo.tm_year + 1900, timeInfo.tm_mon + 1, timeInfo.tm_mday,
      timeInfo.tm_hour, timeInfo.tm_min, timeInfo.tm_sec);
  #endif
  panel.fillScreen(TFT_BLACK);
  panel.setCursor(0, 13);
  panel.setFont(&fonts::Font7); // H32 * V48
  panel.setTextSize(1, 1);    // H32 * V48
  panel.setTextColor(TFT_BLUE);
  panel.printf("%04d-%02d-%02d\n",
    timeInfo.tm_year + 1900, timeInfo.tm_mon + 1, timeInfo.tm_mday);
  panel.setTextSize(1, 2);    // H32 * V96
  panel.setTextColor(TFT_GREEN);
  panel.printf("     %02d:%02d:%02d\n",
    timeInfo.tm_hour, timeInfo.tm_min, timeInfo.tm_sec);
}

//*
//* Audio status functions
//*
void audio_info(const char *info) {
  Serial.print("info        "); Serial.println(info);
}
void audio_id3data(const char *info) { //id3 metadata
  Serial.print("id3data     "); Serial.println(info);
}
void audio_eof_mp3(const char *info) { //end of file
  Serial.print("eof_mp3     "); Serial.println(info);
}
void audio_showstation(const char *info) {
  Serial.print("station     "); Serial.println(info);
  // LovyanGFX
  panel.fillRect (0, 0, 320, 48, TFT_BLACK);  // clear BG 24dot * 2line
  panel.setCursor(0, 0);
  panel.setTextColor(TFT_RED);
  panel.print("Station:");
  panel.setTextColor(TFT_WHITE);
  panel.printf("%s", info); // 16dot * 20column
  panel.drawFastHLine(0, 48, 320, TFT_GREEN);
}
void audio_showstreaminfo(const char *info) {
  Serial.print("streaminfo  "); Serial.println(info);
}
void audio_showstreamtitle(const char *info) {
  Serial.print("streamtitle "); Serial.println(info);
  // LovyanGFX
  panel.fillRect (0, 49, 320, 96, TFT_BLACK); // clear BG 24dot * 4line
  panel.setCursor(0, 49);
  panel.setTextColor(TFT_GREEN);
  panel.print("Title:");
  panel.setTextColor(TFT_WHITE);
  panel.printf("%s", info);
  panel.drawFastHLine(0, 145, 320, TFT_BLUE);
}
void audio_bitrate(const char *info) {
  Serial.print("bitrate     "); Serial.println(info);
}
void audio_commercial(const char *info) { //duration in sec
  Serial.print("commercial  "); Serial.println(info);
}
void audio_icyurl(const char *info) { //homepage
  Serial.print("icyurl      "); Serial.println(info);
}
void audio_lasthost(const char *info) { //stream URL played
  Serial.print("lasthost    "); Serial.println(info);
}
void audio_eof_speech(const char *info) {
  Serial.print("eof_speech  "); Serial.println(info);
}

//*
//* SD Card function
//*
int readStationData(fs::FS &fs, const char * path) {
  char ch;                  // Read character
  int16_t i;                // Station list index
  int16_t j;                // Station name or URL index
  bool StationNameorURLsw;  // StationNameorURLsw(Station name = true、URL = false）

  // Open
  Serial.printf("Reading file: %s\n", path);
  File file = fs.open(path);  // File Open
  // 失敗のときには
  if (!file) {
    Serial.println("Failed to open file for reading");
    return 0;  // 戻る
  }

  // 先頭行を読み込む（#EXTM3U）
  Serial.println("Read the first line from the file: ");  
  while (ch != '\n') {
    ch = file.read();
    Serial.write(ch);
  }

  //  次行を読み込む
  i = 0;  // Station list index initialize
  j = 0;  // Station name or URL index initialize
  StationNameorURLsw = true;  // StationNameorURLsw initialize(Station name = true、URL = false）

  Serial.println("Read from file: ");
  // 現在位置から読み出せるコンテンツのバイト数まで読み込む
  while (file.available()) {
    ch = file.read();
    Serial.write(ch);

    // LFのときには
     if (ch == '\n') {
      Serial.println("LF");
      j = 0;  // Station name or URL index initialize
      StationNameorURLsw = !StationNameorURLsw;  // StationNameorURLsw reversal
      if (StationNameorURLsw) i++;
    
    // それ以外で、CR以外のときには
    } else if (ch != '\r') {
      // 局名のときには
      if (StationNameorURLsw) StationList[i].StationName[j] = ch;
      // それ以外（URL）のときには
      else StationList[i].URL[j] = ch;
      j++;  // Station name or URL index increment
    }
  }
  // 閉じる
  file.close();
  // 戻る
  return (i - 1);
}
// EOF by JG1CCL W3CCL Beard UCHIDA CCLab(Rabbir Hat Design)
